# Copyright (c) 2014 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import contextlib

LOG = None


NAMING_STRATEGY_UUID = 'use_uuid'
NAMING_STRATEGY_NAMES = 'use_name'
NAME_TYPE_TENANT = 'tenant'
NAME_TYPE_NETWORK = 'network'
NAME_TYPE_SUBNET = 'subnet'
NAME_TYPE_PORT = 'port'
NAME_TYPE_ROUTER = 'router'
NAME_TYPE_APP_PROFILE = 'app-profile'
NAME_TYPE_POLICY_TARGET_GROUP = 'policy_target_group'
NAME_TYPE_L3_POLICY = 'l3_policy'
NAME_TYPE_L2_POLICY = 'l2_policy'
NAME_TYPE_POLICY_RULE_SET = 'policy_rule_set'
NAME_TYPE_POLICY_RULE = 'policy_rule'
NAME_TYPE_EXTERNAL_SEGMENT = 'external_segment'
NAME_TYPE_EXTERNAL_POLICY = 'external_policy'
NAME_TYPE_NAT_POOL = 'nat_pool'

NAME_TYPES = set([
    NAME_TYPE_TENANT, NAME_TYPE_NETWORK, NAME_TYPE_SUBNET,
    NAME_TYPE_PORT, NAME_TYPE_ROUTER, NAME_TYPE_APP_PROFILE,
    NAME_TYPE_POLICY_TARGET_GROUP, NAME_TYPE_L3_POLICY,
    NAME_TYPE_L2_POLICY, NAME_TYPE_POLICY_RULE_SET,
    NAME_TYPE_POLICY_RULE, NAME_TYPE_EXTERNAL_SEGMENT,
    NAME_TYPE_EXTERNAL_POLICY, NAME_TYPE_NAT_POOL,
])

MAX_APIC_NAME_LENGTH = 46


@contextlib.contextmanager
def mapper_context(context):
    if context and (not hasattr(context, '_plugin_context') or
                    context._plugin_context is None):
        context._plugin_context = context  # temporary circular reference
        yield context
        context._plugin_context = None     # break circular reference
    else:
        yield context


def truncate(string, max_length):
    if max_length < 0:
        return ''
    return string[:max_length] if len(string) > max_length else string


def _get_conf(conf, name):
    ret = None
    try:
        ret = conf.get(name)
    except AttributeError:
        # ignore if option is not registered
        pass
    return ret


class APICNameMapper(object):

    def __init__(self, db, log, keyclient, keystone_authtoken,
                 strategy=NAMING_STRATEGY_UUID, min_suffix=None,
                 keysession=None, keystoneclientv3=None):
        self.db = db
        self.strategy = strategy
        self.keystone = None
        self.keyclient = keyclient
        self.keysession = keysession
        self.keystone_authtoken = keystone_authtoken
        self.keystoneclientv3 = keystoneclientv3
        self.tenants = {}
        self.min_suffix = min_suffix if min_suffix is not None else 5
        global LOG
        LOG = log.getLogger(__name__)

    @staticmethod
    def get_key_password_params_ext(keystone_conf, suffix=''):
        (auth_url, user, pw, tenant) = \
            APICNameMapper.get_key_password_params(keystone_conf, suffix)
        prj_domain = (_get_conf(keystone_conf, 'project_domain_name') or
                      'Default')
        usr_domain = (_get_conf(keystone_conf, 'user_domain_name') or
                      'Default')
        return auth_url, user, pw, tenant, prj_domain, usr_domain

    @staticmethod
    def get_key_password_params(keystone_conf, suffix=''):
        auth_url = APICNameMapper.get_keystone_url(keystone_conf,
                                                   suffix=suffix)
        user = (_get_conf(keystone_conf, 'username') or
                _get_conf(keystone_conf, 'admin_user'))
        pw = (_get_conf(keystone_conf, 'password') or
              _get_conf(keystone_conf, 'admin_password'))
        tenant = (_get_conf(keystone_conf, 'project_name') or
                  _get_conf(keystone_conf, 'admin_tenant_name'))
        return auth_url, user, pw, tenant

    @staticmethod
    def get_keystone_url(keystone_conf, suffix=''):
        if keystone_conf.get('auth_uri'):
            auth_url = keystone_conf.auth_uri.rstrip('/')
            auth_url = auth_url.split('/')
            if auth_url[-1].startswith('v2') or auth_url[-1].startswith('v3'):
                auth_url = '/'.join(auth_url[:-1])
            else:
                auth_url = '/'.join(auth_url)
            if not auth_url.endswith(suffix):
                auth_url += suffix
        else:
            auth_url = ('%s://%s:%s' % (
                keystone_conf.auth_protocol,
                keystone_conf.auth_host,
                keystone_conf.auth_port) + suffix)
        return auth_url + '/'

    def mapper(name_type):
        """Wrapper to land all the common operations between mappers."""
        def wrap(func):
            def inner(inst, context, resource_id, remap=False, prefix='',
                      existing=False):
                if existing:
                    return ApicName(resource_id, resource_id, context, inst,
                                    func.__name__, existing=True)
                if remap:
                    inst.db.delete_apic_name(resource_id)
                else:
                    saved_name = inst.db.get_apic_name(resource_id,
                                                       name_type)
                    if saved_name:
                        result = saved_name[0]
                        if prefix:
                            result = prefix + result
                            result = truncate(result, MAX_APIC_NAME_LENGTH)
                        return ApicName(result, resource_id, context,
                                        inst, func.__name__, prefix=prefix)
                name = ''
                try:
                    name = func(inst, context, resource_id)
                except Exception as e:
                    LOG.warn(("Exception in looking up name %s"), name_type)
                    LOG.error(e.message)

                purged_id = re.sub(r"-+", "-", resource_id)
                result = purged_id[:inst.min_suffix]
                if name:
                    name = re.sub(r"-+", "-", name)
                    if inst.strategy == NAMING_STRATEGY_NAMES:
                        result = name
                    elif inst.strategy == NAMING_STRATEGY_UUID:
                        # Keep as many uuid chars as possible
                        id_suffix = "_" + result
                        max_name_length = MAX_APIC_NAME_LENGTH - len(id_suffix)
                        result = truncate(name, max_name_length) + id_suffix

                    result = truncate(result, MAX_APIC_NAME_LENGTH)
                    # Remove forbidden whitespaces
                    result = result.replace(' ', '')
                    if inst.strategy == NAMING_STRATEGY_UUID:
                        result = inst._grow_id_if_needed(
                            purged_id, name_type, result,
                            start=inst.min_suffix)
                else:
                    result = purged_id

                inst.db.update_apic_name(resource_id, name_type, result)
                if prefix:
                    result = prefix + result
                    result = truncate(result, MAX_APIC_NAME_LENGTH)
                return ApicName(result, resource_id, context, inst,
                                func.__name__, prefix=prefix)
            return inner
        return wrap

    def _grow_id_if_needed(self, resource_id, name_type, current_result,
                           start=0):
        result = current_result
        if result.endswith('_'):
            result = result[:-1]
        try:
            x = 0
            while True:
                if self.db.get_filtered_apic_names(neutron_type=name_type,
                                                   apic_name=result):
                    if x == 0 and start == 0:
                        result += '_'
                    # This name overlaps, add more ID characters
                    result += resource_id[start + x]
                    x += 1
                else:
                    break
        except AttributeError:
            LOG.info("Current DB API doesn't support "
                     "get_filtered_apic_names.")
        except IndexError:
            LOG.debug("Ran out of ID characters.")
        return result

    def get_tenant_name(self, tenant_id, require_keystone_session=False):
        tenant_name = None
        if tenant_id in self.tenants:
            tenant_name = self.tenants.get(tenant_id)
        else:
            if self.keystone is None:
                try:
                    self.keystone = self._get_keystone()
                except Exception as ex:
                    LOG.exception(ex)
                    if require_keystone_session:
                        return None
                    else:
                        return tenant_id
            try:
                # v2 API
                LOG.debug("Calling tenant API")
                tenant_list = self.keystone.tenants.list()
            except AttributeError:
                # v3 API
                LOG.debug("Calling project API")
                tenant_list = self.keystone.projects.list()

            LOG.debug("Tenant list received %s" % tenant_list)
            for tenant in tenant_list:
                self.tenants[tenant.id] = tenant.name
                if tenant.id == tenant_id:
                    tenant_name = tenant.name
        return tenant_name

    def update_tenant_name(self, tenant_id):
        if self.keystone is None:
            try:
                self.keystone = self._get_keystone()
            except Exception as ex:
                LOG.exception(ex)
        if self.keystone:
            try:
                # v2 API
                LOG.debug("Calling tenant API")
                tenant = self.keystone.tenants.get(tenant_id)
            except AttributeError:
                # v3 API
                LOG.debug("Calling project API")
                tenant = self.keystone.projects.get(tenant_id)
            # only return tenant name when there is a change
            if tenant and self.tenants.get(tenant_id) != tenant.name:
                self.tenants[tenant.id] = tenant.name
                return tenant.name
        return None

    def is_tenant_in_apic(self, tenant_id):
        if self.db.get_apic_name(tenant_id, NAME_TYPE_TENANT):
            return True
        return False

    @mapper(NAME_TYPE_TENANT)
    def tenant(self, context, tenant_id):
        return self.get_tenant_name(tenant_id)

    @mapper(NAME_TYPE_NETWORK)
    def network(self, context, network_id):
        network = context._plugin.get_network(
            context._plugin_context, network_id)
        network_name = network['name']
        return network_name

    @mapper(NAME_TYPE_SUBNET)
    def subnet(self, context, subnet_id):
        subnet = context._plugin.get_subnet(context._plugin_context, subnet_id)
        subnet_name = subnet['name']
        return subnet_name

    @mapper(NAME_TYPE_PORT)
    def port(self, context, port_id):
        port = context._plugin.get_port(context._plugin_context, port_id)
        port_name = port['name']
        return port_name

    @mapper(NAME_TYPE_ROUTER)
    def router(self, context, router_id):
        return context._plugin_context.session.execute(
            'SELECT * from routers WHERE id = :id',
            {'id': router_id}).fetchone().name

    @mapper(NAME_TYPE_POLICY_TARGET_GROUP)
    def policy_target_group(self, context, policy_target_group_id):
        epg = context._plugin.get_policy_target_group(context._plugin_context,
                                                 policy_target_group_id)
        return epg['name']

    @mapper(NAME_TYPE_L3_POLICY)
    def l3_policy(self, context, l3_policy_id):
        l3_policy = context._plugin.get_l3_policy(context._plugin_context,
                                                  l3_policy_id)
        return l3_policy['name']

    @mapper(NAME_TYPE_L2_POLICY)
    def l2_policy(self, context, l2_policy_id):
        l2_policy = context._plugin.get_l2_policy(context._plugin_context,
                                                  l2_policy_id)
        return l2_policy['name']

    @mapper(NAME_TYPE_POLICY_RULE_SET)
    def policy_rule_set(self, context, policy_rule_set_id):
        policy_rule_set = context._plugin.get_policy_rule_set(
            context._plugin_context, policy_rule_set_id)
        return policy_rule_set['name']

    @mapper(NAME_TYPE_POLICY_RULE)
    def policy_rule(self, context, policy_rule_id):
        policy_rule = context._plugin.get_policy_rule(context._plugin_context,
                                                      policy_rule_id)
        return policy_rule['name']

    @mapper(NAME_TYPE_EXTERNAL_SEGMENT)
    def external_segment(self, context, external_segment_id):
        external_segment = context._plugin.get_external_segment(
            context._plugin_context, external_segment_id)
        return external_segment['name']

    @mapper(NAME_TYPE_EXTERNAL_POLICY)
    def external_policy(self, context, external_policy_id):
        external_policy = context._plugin.get_external_policy(
            context._plugin_context, external_policy_id)
        return external_policy['name']

    @mapper(NAME_TYPE_NAT_POOL)
    def nat_pool(self, context, nat_pool_id):
        nat_pool = context._plugin.get_nat_pool(context._plugin_context,
                                                nat_pool_id)
        return nat_pool['name']

    def pre_existing(self, context, object_id):
        return ApicName(object_id, object_id, context, self,
                        self.pre_existing.__name__, existing=True)

    def echo(self, context, object_id):
        return ApicName(object_id, object_id, context, self,
                        self.echo.__name__)

    def app_profile(self, context, app_profile, remap=False):
        if remap:
            self.db.delete_apic_name('app_profile')
        # Check if a profile is already been used
        saved_name = self.db.get_apic_name('app_profile',
                                           NAME_TYPE_APP_PROFILE)
        if not saved_name:
            self.db.update_apic_name('app_profile', NAME_TYPE_APP_PROFILE,
                                     app_profile)
            result = app_profile
        else:
            result = saved_name[0]
        return ApicName(result, app_profile, None,
                        self, self.app_profile.__name__)

    def delete_apic_name(self, object_id):
        self.db.delete_apic_name(object_id)

    def is_valid_name_type(self, name_type):
        return name_type in NAME_TYPES

    def _get_keystone(self):
        if self.keystoneclientv3:
            return self.keystoneclientv3

        keystone_conf = self.keystone_authtoken
        if not self.keysession:
            # This assumes v2
            LOG.info("No session provided, assuming Keystone v2 API")
            auth_url, user, pw, tenant = self.get_key_password_params(
                keystone_conf, suffix='/v2.0')
            return self.keyclient.Client(
                username=user, password=pw, tenant_name=tenant,
                auth_url=auth_url)
        else:
            auth_url = (keystone_conf.get('identity_uri') or
                        self.get_keystone_url(keystone_conf))
            LOG.info("Keystone identity URI %s" % auth_url)
            version = keystone_conf.get('auth_version')
            if version and version.startswith('v'):
                version = version.lstrip('v')

            version = (2,) if not version else (int(version),)
            LOG.info("Keystone v%s API provided with session" % str(version))
            return self.keyclient.Client(
                session=self.keysession, endpoint=auth_url, version=version)


class ApicName(object):

    def __init__(self, mapped, uid='', context=None, inst=None, fname='',
                 prefix='', existing=False):
        self.uid = uid
        self.context = context
        self.inst = inst
        self.fname = fname
        self.value = mapped
        self.prefix = prefix
        self.existing = existing

    def renew(self):
        if self.uid and self.inst and self.fname:
            # temporary circular reference
            with mapper_context(self.context) as ctx:
                result = getattr(self.inst, self.fname)(
                    ctx, self.uid, remap=True, prefix=self.prefix)
            self.value = result.value
            return self

    def __str__(self):
        return self.value
